%% Demo for comparison of the following algorithms:
%% CD-BM3D: algorithm based on complex-domain sparsification of complex valued noisy image (HOSVD is applied to complex-valued variables);
%% ImRe-BM3D: algorithm based on real-valued sparsification of complex valued noisy image (HOSVD is applied to 4D groups of real valued variables);

%% Paper submited to Signal Processing
%% 'Nonlocal block-matching sparsity for complex-valued data denoising'
%% Vladimir Katkovnik, Nikolay Ponomarenko and Karen Egiazarian, December, 2016

%% Vladimir Katkovnik, Tampere, TUT, December 2016

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

close all
clear all
addpath(genpath('PUMA'));
addpath(genpath('hosvddir'));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% showing options
if_figure_on=1;    %% show figures
if_comments=1;     %% show text results

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% selection of parameters of the demo

denoising_Method=0;
while (denoising_Method<1)|(denoising_Method>2)
  disp(' ');
  denoising_Method = input('Select denoising method (1 - CD-BM3D, 2 - ImRe-BM3D): ');
end

test_Phase=0;
while (test_Phase<1)|(test_Phase>5)
  disp(' ');
  disp('Test phase images:');
  disp('1 - Lena image');
  disp('2 - Cameraman image');
  disp('3 - Truncated Gaussian distribution');
  disp('4 - Mountains image');
  disp('5 - Lena plus quadratic phase');
  test_Phase = input('Select test phase image (1-5): ');
end

test_Amplitude=0;
while (test_Amplitude<1)|(test_Amplitude>3)
  disp(' ');
  disp('Test amplitude images:');
  disp('1 - Invariant amplitude');
  disp('2 - Normalized absolute phase');
  disp('3 - Cosine of phase multiplied by 15');
  test_Amplitude = input('Select test amplitude image (1-3): ');
end

disp(' ');
k0 = input('Enter k0 factor (see equation 28, default value is 1): ');

disp(' ');
k1 = input('Enter k1 factor (see equation 28, default value is 1): ');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% test complex valued image 

switch(test_Phase)
   case 1
     varphi=double(imread('image_Lena256.png'))/255*pi/2;
     PhaseName='Lena';
   case 2 
     varphi=double(imread('image_Cameraman256.png'))/255*pi/2;
     PhaseName='Cameraman';
   case 3               %% Gaussian abs phase
     varphi=fspecial('gaussian',256,50);
     varphi(1:128,1:128)=0;
     tmp1=max(varphi(:));
     tmp2=min(varphi(:));
     varphi=(varphi-tmp2)/(tmp1-tmp2)*44;
     PhaseName='Gaussian'; 
   case 4       %%  Mountains  (EXP 4)
     M = 256; N = 256;
     S=2055615866; randn('seed', S);  
     varphi=randn(M,N);
     fc = 0.01; % define cut frequency
     mask = zeros(M,N);
     mask(round(M/2-fc*M/2):round(M/2+fc*M/2), ...
          round(N/2-fc*N/2):round(N/2+fc*N/2)) = 1;
     varphi =  2400* real(ifft2(fftshift(fftshift(fft2(varphi)).*mask)));
     PhaseName='Mountains'; 
   case 5
     varphi=double(imread('image_Lena256.png'))/255*pi/2;
     M=256; N=256;
     [X,Y] = meshgrid(-round(N/2): -round(N/2)+N-1, ...
                      -round(M/2): -round(M/2)+M-1);
     quadratic_phase=-(X.^2+Y.^2)/(4*(M/2)^2);    %% Quadratic phase     
     quadratic_phase=quadratic_phase+max(abs(quadratic_phase(:)));    
     quadratic_phase=quadratic_phase*50; 
     varphi=varphi+quadratic_phase;
     PhaseName='Quadratic'; 
end

[M,N]=size(varphi);

switch(test_Amplitude)
   case 1
     B=ones(size(varphi))*k1+k0; 
     AmplitudeName='Invariant';
   case 2
     B=abs(varphi/max(varphi(:)))*k1+k0;  
     AmplitudeName='Normalized';
   case 3
     B=abs(cos(15*varphi));
     B=B-min(min(B));
     B=B*k1+k0;                  
     AmplitudeName='Cosine';
end

Y=B.*exp(1i*varphi);         %% True complex signal
varphi_wrapped=angle(Y);     %% phase wrapping

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

sigma_set= [.05 .1 .2 .5 .9 ];           %% Set of sigma values for the noise in phase component
lambda_CD_BM3D = [1 1.1 1.2 1.25 1.25];  %% threshold parameter for CD-BM3D for different sigmas (Figure 9)
lambda_ImRe_BM3D = [.75 .77 .8 .85 .9];  %% threshold parameter for ImRe-BM3D for different sigmas (Figure 9)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%% MAIN PROGRAM %%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for n=1:length(sigma_set)               %% Loop over sigma of noise
  
  sigma = sigma_set(n);           %% sigma for phase
  sigma_ampl= sigma*mean(B(:));   %% sigma for amplitude
  if if_comments
    disp(' ');
    disp(['sigma = ', num2str(sigma)]);
  end

  %%%%%%%%% Noisy Gaussian Data %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  S=2055615866; randn('seed', S);
  Z=Y + sigma_ampl*(randn(size(B))+1i*randn(size(B)));  %% Noisy signal


  %% parameters of denoising

  NStep=3;                  %% NStep: sliding step (usually equal to 3) to process every next reference block
  nmFact=1;                 %% nmFact: multiplication factor (usually equal to 1) for maximum number of similar blocks (maximum size of the 3rd dimension of a 3D array)
  Ns=39;                    %% Ns: length of the side of the search neighborhood for full-search block-matching (BM), must be odd (ususlly equal to 39)
  ConfFact=2;               %% ConfFact: parameter (usually equal to 2) for the two sides symmetric confidence interval for grouping
  upConfFact=10;            %% upConfFact: parameter (usually equal to 10) for the upper bound in the two sides non-symmetric confidence interval for grouping
  lambda2D=0;               %% lambda2D: threshold parameter (usually equal to 0) for the coarse initial denoising used in the d-distance measure
  SVD3D=1;                  %% SVD3D: 1 stays for high-order 3D SVD, otherwise 2D SVD
  
  if denoising_Method==1     %% CD-BM3D
  
    MethodName='CD-BM3D';
    
    if if_comments
      fprintf('\nCD-BM3D denoising... ');
      tic;
    end
    
    lambda=lambda_CD_BM3D(n); %% threshold parameter (1.5 .. 3)

    CDBM3D_hat = CD_BM3D_Separate_Thresholding_parameters(sigma_ampl, Z, lambda, NStep, nmFact, Ns, ConfFact, upConfFact, lambda2D, SVD3D);

    if if_comments
      disp('done.');
      toc;
      fprintf('Wiener filtering... ');
      tic;
    end

    CDBM3D_Wiener_hat = CD_BM3D_Separate_Thresholding_wiener_parameters(sigma_ampl, Z, CDBM3D_hat, lambda, NStep, nmFact, Ns, ConfFact, upConfFact, lambda2D, SVD3D);


  elseif denoising_Method==2     %% ImRe-BM3D

    MethodName='ImRe-BM3D';
    
    if if_comments
      fprintf('\nImRe-BM3D denoising... ');
      tic;
    end
    
    lambda=lambda_ImRe_BM3D(n); %% threshold parameter (1.5 .. 3)

    CDBM3D_hat = ImRe_BM3D_parameters(sigma_ampl, Z, lambda, NStep, nmFact, Ns, ConfFact, upConfFact, lambda2D, SVD3D);

    if if_comments
      disp('done.');
      toc;
      fprintf('Wiener filtering... ');
      tic;
    end

    CDBM3D_Wiener_hat = ImRe_BM3D_wiener_parameters(sigma_ampl, Z, CDBM3D_hat, lambda, NStep, nmFact, Ns, ConfFact, upConfFact, lambda2D, SVD3D);
  end
    
  if if_comments
    disp('done.');
    toc;
  end

  filtered_phase_wrapped=angle(CDBM3D_Wiener_hat);
  filtered_amplitude=abs(CDBM3D_Wiener_hat);

  %% estimation of absolute phase
  
  potential.quantized = 'no';
  potential.threshold = 0.5;
  verbose='no';
  filtered_phase_puma = puma_ho(filtered_phase_wrapped,0.5, 'potential',potential,'verbose', verbose); 

  %% metrics calculation
  
  delt = (wrap(varphi - filtered_phase_puma)).^2;
  RMSE_ABS_PHASE = mean(delt(:)).^0.5;                     %% RMSE for absolute phase

  delt = (B - abs(Z)).^2;
  PSNR_NOISY_AMPL = 10*log10(max(B(:)).^2/mean(delt(:)));  %% PSNR for noisy amplitude

  delt = (B - filtered_amplitude).^2;
  PSNR_AMPL = 10*log10(max(B(:)).^2/mean(delt(:)));        %% PSNR for amplitude

  delt = (varphi_wrapped - angle(Z)).^2;
  PSNR_NOISY_PHASE = 10*log10(4*pi^2/mean(delt(:)));       %% PSNR for noisy wrapped phase

  delt = (varphi_wrapped - filtered_phase_wrapped).^2;
  PSNR_PHASE = 10*log10(4*pi^2/mean(delt(:)));             %% PSNR for wrapped phase

  if if_comments
    fprintf('\nPSNR_PHASE = %.1f dB, RMSE_ABS_PHASE = %.3f, PSNR_AMPL = %.1f dB\n',PSNR_PHASE,RMSE_ABS_PHASE,PSNR_AMPL);
  end

  if if_figure_on==1
    figure; 
    set(gcf,'name',[MethodName ', Phase: ' PhaseName ', Amplitude: ' AmplitudeName ', Sigma = ' num2str(sigma)],'numbertitle','off') 
    
    subplot(2,3,1); 
    imshow(angle(Z),[]); 
    title(['Noisy WRAP PHASE, PSNR = ', num2str(PSNR_NOISY_PHASE,3) ' dB']);

    subplot(2,3,2);
    imshow(abs(Z),[]);
    title(['Noisy AMPL, PSNR = ', num2str(PSNR_NOISY_AMPL,3), ' dB']);

    subplot(2,3,3); 
    imshow(filtered_phase_puma,[]); 
    title(['ABS PHASE, RMSE = ', num2str(RMSE_ABS_PHASE,3)]);

    subplot(2,3,4);
    imshow(filtered_phase_wrapped,[]);
    title(['WRAP PHASE, PSNR = ', num2str(PSNR_PHASE,3), ' dB']);

    subplot(2,3,5);
    imshow(filtered_amplitude,[]);
    title(['AMPL, PSNR = ', num2str(PSNR_AMPL,3), ' dB']);
    
    subplot(2,3,6);
    imshow(wrap(filtered_phase_puma-varphi),[]);
    title(['WRAP PHASE ERRORS, sigma = ', num2str(sigma)]);

    if (test_Phase>2)&&(test_Phase<5)    %% wrapped phase

      %% noisy wrapped phase after processing by PUMA script
      phase_noisy=puma_ho(angle(Z),0.5, 'potential',potential,'verbose', verbose);  
      delt = (wrap(varphi - phase_noisy)).^2;
      RMSE_NOISY_PHASE = mean(delt(:)).^0.5;         %% RMSE for noisy absolute phase
      
      ind=1:256;

      figure; 
      set(gcf,'name',['3D representation, ' MethodName ', Phase: ' PhaseName ', Amplitude: ' AmplitudeName ', Sigma = ' num2str(sigma)],'numbertitle','off') 
      subplot(1,2,1); 
      surfl(ind,ind,phase_noisy); shading interp; colormap(gray);
      axis off;
      title(['NOISY ABS PHASE, RMSE = ', num2str(RMSE_NOISY_PHASE,3)]);
    
      subplot(1,2,2); 
      surfl(ind,ind,filtered_phase_puma); shading interp; colormap(gray);
      axis off;
      title(['ABS PHASE, RMSE = ', num2str(RMSE_ABS_PHASE,3)]);

    end
  end

end
